/* gprs_coding_scheme.h
 *
 * Copyright (C) 2015 by Sysmocom s.f.m.c. GmbH
 * Author: Jacob Erlbeck <jerlbeck@sysmocom.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#pragma once

#include <stdint.h>
#include <stddef.h>

#define MAX_NUM_ARQ           2      /* max. number of ARQ */
#define MAX_NUM_MCS           9     /* max. number of MCS */
#define MAX_NUM_PS            3      /* max. number of puncturing schemes */

#define EGPRS_ARQ1            0x0
#define EGPRS_ARQ2            0x1

struct egprs_ps_sel_tbl{
	   	uint8_t next_ps;              /* next puncturing scheme */
		uint8_t action;              /* action to be taken on counter */
	};


enum PsTypes{
		EGPRS_PS_TYPE_2,
		EGPRS_PS_TYPE_3,
	};

	enum PsValues {
		EGPRS_PS1 = 0, 
		EGPRS_PS2, 
		EGPRS_PS3,
	};

class GprsCodingScheme {
public:
	enum Scheme {
		UNKNOWN,
		CS1, CS2, CS3, CS4,
		MCS1,MCS2, MCS3, MCS4,
		MCS5,MCS6, MCS7, MCS8, MCS9,
		NUM_SCHEMES
	};

	enum Mode {
		GPRS,
		EGPRS,
	};

	enum HeaderType {
		HEADER_INVALID,
		HEADER_GPRS_CONTROL,
		HEADER_GPRS_DATA,
		HEADER_EGPRS_DATA_TYPE_1,
		HEADER_EGPRS_DATA_TYPE_2,
		HEADER_EGPRS_DATA_TYPE_3,
	};
	
	GprsCodingScheme(Scheme s = UNKNOWN,Mode scheme_type = GPRS);

	operator bool() const {return m_scheme != UNKNOWN;}
	operator int()  const {return (int)m_scheme;}
	void operator =(Scheme s);
	GprsCodingScheme operator + (Scheme scheme);
	GprsCodingScheme operator -(Scheme scheme);
	void operator =(GprsCodingScheme o);
	void operator =(int o);
	bool operator ==(Scheme o);
	bool operator >= (Scheme o);
	bool isValid()   const {return UNKNOWN <= m_scheme && m_scheme <= MCS9;}
	bool isGprs()   const {return CS1 <= m_scheme && m_scheme <= CS4;}
	bool isEgprs()  const {return m_scheme >= MCS1;}
	bool isCompatible(Mode mode) const;
	uint8_t get_cps(uint8_t ps, bool is_mcs8_retx = false);

	unsigned int sizeUL() const;
	unsigned int sizeDL() const;
	unsigned int maxBytesUL() const;
	unsigned int maxBytesDL() const;
	unsigned int spareBitsUL() const;
	unsigned int spareBitsDL() const;
	unsigned int maxDataBlockBytes() const;
	unsigned int numDataBlocks() const;
	Scheme coding_scheme() const;
	const char *name() const;
	HeaderType headerTypeData() const;
	HeaderType headerTypeControl() const;

	static GprsCodingScheme getBySizeUL(unsigned size);
	Scheme get_retx_mcs(GprsCodingScheme *current_mcs) const;

private:
	enum Scheme m_scheme;
	enum Mode   m_scheme_type;
};

struct egprs_mcs_ps{
   		GprsCodingScheme::Scheme mcs;   /* GprsCodingScheme::MCS */           
		uint8_t ps;                     /* MEAN_BEP */
};

inline bool GprsCodingScheme::isCompatible(Mode mode) const
{
	switch (mode) {
	case GPRS: return isGprs();
	case EGPRS: return isEgprs();
	}

	return false;
}

inline GprsCodingScheme::HeaderType GprsCodingScheme::headerTypeControl() const
{
	return HEADER_GPRS_CONTROL;
}

inline GprsCodingScheme::GprsCodingScheme(Scheme s, Mode scheme_type )
	: m_scheme(s), m_scheme_type(scheme_type)
{
	if (!isValid())
		m_scheme = UNKNOWN;
}

inline void GprsCodingScheme::operator =(Scheme s)
{
	m_scheme = s;
	if (!isValid())
		m_scheme = UNKNOWN;
}

inline void GprsCodingScheme::operator =(GprsCodingScheme o)
{
	m_scheme = o.m_scheme;
	m_scheme_type = o.m_scheme_type;
}
inline bool GprsCodingScheme::operator ==(Scheme o)
{
	if( this->m_scheme != o)
	{
		return false;
	}
	return true;
}

inline GprsCodingScheme GprsCodingScheme::operator + (Scheme o)
{
	GprsCodingScheme cs((Scheme)(this->m_scheme + o));
	return	(cs);
}

inline GprsCodingScheme GprsCodingScheme::operator - (Scheme o)
{
	GprsCodingScheme       cs((Scheme)(this->m_scheme - o));
	return	(cs);
}

inline bool GprsCodingScheme::operator >= (Scheme o)
{
	if(this->m_scheme >= o){
		return true;
	}
	return false;
}
